/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2010 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>

#include "config.h"
#include "ctunnel.h"

#ifdef HAVE_OPENSSL
crypto_ctx *openssl_encrypt_init(struct options opt)
{
    const EVP_CIPHER *c = NULL;
    crypto_ctx *ctx = calloc(1, sizeof(crypto_ctx));

    /* STREAM CIPHERS ONLY */
    OpenSSL_add_all_ciphers();
    c = EVP_get_cipherbyname(opt.key.cipher);
    if (!c) {
	syslog(LOG_CRIT, "Cipher \"%s\" not found!", opt.key.cipher);
        fprintf(stdout, "Cipher \"%s\" not found!\n", opt.key.cipher);
        exit(1);
    }
    EVP_CIPHER_CTX_init(ctx);
    EVP_EncryptInit_ex(ctx, EVP_get_cipherbyname(opt.key.cipher), NULL,
                       opt.key.key, opt.key.iv);
    EVP_CIPHER_CTX_set_padding(ctx, 1);

    return ctx;

}
crypto_ctx *openssl_decrypt_init(struct options opt)
{
    const EVP_CIPHER *c = NULL;
    crypto_ctx *ctx = calloc(1, sizeof(crypto_ctx));

    /* STREAM CIPHERS ONLY */
    OpenSSL_add_all_ciphers();
    c = EVP_get_cipherbyname(opt.key.cipher);
    if (!c) {
	syslog(LOG_CRIT, "Cipher \"%s\" not found!", opt.key.cipher);
        fprintf(stdout, "Cipher \"%s\" not found!\n", opt.key.cipher);
        exit(1);
    }

    EVP_CIPHER_CTX_init(ctx);
    EVP_DecryptInit_ex(ctx, EVP_get_cipherbyname(opt.key.cipher), NULL,
                       opt.key.key, opt.key.iv);
    EVP_CIPHER_CTX_set_padding(ctx, 1);

    return ctx;
}
struct Packet openssl_do_encrypt(crypto_ctx *ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv)
{
    struct Packet crypto;

    crypto.len = 0;
    crypto.data = calloc(1, size);

    if (!EVP_EncryptUpdate(ctx, crypto.data, &crypto.len, intext, size)) {
	syslog(LOG_CRIT, "EVP_EncryptUpdate: Error");
	fprintf(stderr, "EVP_EncryptUpdate: Error\n");
    }

    return crypto;
}
struct Packet openssl_do_decrypt(crypto_ctx *ctx, unsigned char *intext,
                                 int size, unsigned char *key,
				 unsigned char *iv)
{
    struct Packet crypto;

    crypto.len = 0;
    crypto.data = calloc(1, size);

    if (!EVP_DecryptUpdate(ctx, crypto.data, &crypto.len, intext, size)) {
	syslog(LOG_CRIT, "EVP_DecryptUpdate: Error");
	fprintf(stderr, "EVP_DecryptUpdate: Error\n");
    }

    return crypto;
}
#else
crypto_ctx gcrypt_crypto_init(struct options opt)
{
    int ret = 0;
    int mode;

    if (!strcmp(opt.key.mode, "ecb"))
	mode = GCRY_CIPHER_MODE_ECB;
    if (!strcmp(opt.key.mode, "cfb"))
	mode = GCRY_CIPHER_MODE_CFB;
    if (!strcmp(opt.key.mode, "cbc"))
	mode = GCRY_CIPHER_MODE_CBC;
    if (!strcmp(opt.key.mode, "ofb"))
	mode = GCRY_CIPHER_MODE_OFB;

    crypto_ctx ctx;
    ret = gcry_cipher_open(&ctx, gcry_cipher_map_name(opt.key.cipher), mode, 0);
    if (!ctx) {
	syslog(LOG_CRIT, "gcry_open_cipher: %s", gpg_strerror(ret));
        fprintf(stdout, "gcry_open_cipher: %s\n", gpg_strerror(ret));
        exit(ret);
    }
    ret = gcry_cipher_setkey(ctx, opt.key.key, 16);
    if (ret) {
	syslog(LOG_CRIT, "gcry_cipher_setkey: %s", gpg_strerror(ret));
        fprintf(stdout, "gcry_cipher_setkey: %s\n", gpg_strerror(ret));
        exit(ret);
    }
    ret = gcry_cipher_setiv(ctx, opt.key.iv, 16);
    if (ret) {
	syslog(LOG_CRIT, "gcry_cipher_setiv: %s", gpg_strerror(ret));
        fprintf(stdout, "gcry_cipher_setiv: %s\n", gpg_strerror(ret));
        exit(ret);
    }
    return ctx;
}
struct Packet gcrypt_do_encrypt(crypto_ctx ctx, unsigned char *intext,
	                        int size, unsigned char *key,
				unsigned char *iv)
{
    int ret = 0;
    struct Packet crypto;

    crypto.len = size;
    crypto.data = calloc(1, crypto.len);
    ret = gcry_cipher_encrypt(ctx, crypto.data, crypto.len,
	                      intext, size);
    if (ret) {
	syslog(LOG_CRIT, "Encrypt Error: %s", gpg_strerror(ret));
	fprintf(stderr, "Encrypt Error: %s\n", gpg_strerror(ret));
    }

    return crypto;
}
struct Packet gcrypt_do_decrypt(crypto_ctx ctx, unsigned char *intext,
	                        int size, unsigned char *key,
				unsigned char *iv)
{
    int ret = 0;
    struct Packet crypto;

    crypto.len = size;
    crypto.data = calloc(1, crypto.len);
    ret = gcry_cipher_decrypt(ctx, crypto.data,
	                      crypto.len, intext, size);

    return crypto;
}
#endif
