/*
 * Ctunnel - Cyptographic tunnel.
 * Copyright (C) 2008-2010 Jess Mahan <ctunnel@nardcore.org>
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <time.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/select.h>
#include <sys/wait.h>
#include <string.h>
#include <ctype.h>
#include <errno.h>

#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <netinet/in_systm.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/ioctl.h>

#include "config.h"
#include "ctunnel.h"
#include "comp.h"
#include "crypto.h"
#include "net.h"
#include "opt.h"

void crypto_shutdown(int sig)
{
    syslog(LOG_INFO, "Shutting Down\n");
    fprintf(stdout, "[ctunnel] Shutting Down\n");
#ifdef HAVE_OPENSSL
    EVP_cleanup();
#endif
    exit(0);
}

int main(int argc, char *argv[])
{
    int srv_sockfd = 0;
    fd_set rfd;
    int ret = 0;
    struct sockaddr_in pin;
    socklen_t addrsize = 0;
    void *data = NULL;
    pid_t pid;
    struct options opt;
    struct Packet pkt_in = { NULL, 0 }, pkt_out = { NULL, 0 };
#ifdef HAVE_OPENSSL
    crypto_ctx *ectx;
    crypto_ctx *dctx;
    struct Packet (*do_encrypt) (crypto_ctx * ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_encrypt = openssl_do_encrypt;
    struct Packet (*do_decrypt) (crypto_ctx * ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_decrypt = openssl_do_decrypt;
#else
    crypto_ctx ectx = NULL;
    crypto_ctx dctx = NULL;
    struct Packet (*do_encrypt) (crypto_ctx ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_encrypt = gcrypt_do_encrypt;
    struct Packet (*do_decrypt) (crypto_ctx ctx, unsigned char *intext,
				 int size, unsigned char *key,
				 unsigned char *iv);
    do_decrypt = gcrypt_do_decrypt;
#endif
    struct Network *net_cli = NULL, *net_srv = NULL;
    struct Compress comp;


    fprintf(stdout, "%s %d.%d %s\n", SOFTWARE_NAME, VERSION_MAJOR,
	                             VERSION_MINOR, COPYRIGHT);

    openlog("ctunnel", LOG_NDELAY | LOG_PID, LOG_DAEMON);
    syslog(LOG_INFO, "ctunnel %d.%d starting", VERSION_MAJOR, VERSION_MINOR);

    memset(opt.host, 0x00, sizeof(opt.host));
    opt = get_options(argc, argv);

    if (opt.comp == 1) { 
	fprintf(stdout, "libz Compression Enabled\n");
	syslog(LOG_INFO, "libz Compression Enabled\n");
	comp = z_compress_init(opt);
    }

#ifdef HAVE_OPENSSL
    ectx = openssl_encrypt_init(opt);
    dctx = openssl_decrypt_init(opt);
#else
    ectx = gcrypt_crypto_init(opt);
    dctx = gcrypt_crypto_init(opt);
#endif


    signal(SIGCHLD, SIG_IGN);
    signal(SIGINT, crypto_shutdown);
    signal(SIGTERM, crypto_shutdown);

    if (opt.proto == 0) {
	srv_sockfd = tcp_listen(opt.port_listen);
	net_cli = calloc(1, sizeof(struct Network));
	net_srv = calloc(1, sizeof(struct Network));
	fprintf(stdout, "TCP ");
    } else {
	net_srv = udp_bind(opt.port_listen);
	net_cli = udp_connect(opt.host, opt.port_forward);
	data = calloc(1, sizeof(char *) * PACKET_SIZE);
	fprintf(stdout, "UDP ");
    }
#ifdef HAVE_OPENSSL
    fprintf(stdout, "[OpenSSL] Listening on %d using %s\n",
	    opt.port_listen, opt.key.cipher);
#else
    fprintf(stdout, "[gcrypt] Listening on %d using %s %s\n",
	    opt.port_listen, opt.key.cipher, opt.key.mode);
#endif

    if (opt.proto == 0) {
#ifdef HAVE_OPENSSL
	syslog(LOG_INFO, "TCP [OpenSSL] Listening on %d using %s",
	       opt.port_listen, opt.key.cipher);
#else
	syslog(LOG_INFO, "TCP [libgcrypt] Listening on %d using %s %s",
	       opt.port_listen, opt.key.cipher, opt.key.mode);
#endif
    } else {
#ifdef HAVE_OPENSSL
	syslog(LOG_INFO, "UDP [OpenSSL] Listening on %d using %s",
	       opt.port_listen, opt.key.cipher);
#else
	syslog(LOG_INFO, "UDP [libgcrypt] Listening on %d using %s %s",
	       opt.port_listen, opt.key.cipher, opt.key.mode);
#endif
    }

    if (srv_sockfd < 0) {
	syslog(LOG_CRIT, "listen() %s\n", strerror(errno));
	fprintf(stderr, "listen() %s\n", strerror(errno));
	exit(1);
    }

    if (opt.daemon != 1) {
	syslog(LOG_INFO, "Going into background\n");
	fprintf(stdout, "Going into background\n");
	pid = fork();
	if (pid == -1) {
	    syslog(LOG_CRIT, "fork() %s\n", strerror(errno));
	    fprintf(stdout, "ERROR: %s\n", strerror(errno));
	    exit(1);
	}
	if (pid > 0) {
	    exit(0);
	}
    }

    while (1) {
	if (opt.proto == 0) {
	    net_srv->sockfd = accept(srv_sockfd, (struct sockaddr *) &pin,
		                     &addrsize);
	    pid = fork();
	} else {
	    pid = 0;
	}

	if (pid == 0) {
	    if (opt.proto == 0) {
		close(srv_sockfd);
		net_cli->sockfd = tcp_connect(opt.host, opt.port_forward);
	    }

	    data = calloc(1, sizeof(char *) * PACKET_SIZE);

	    while (1) {
		if (opt.proto == 1) {
#ifdef HAVE_OPENSSL
		    EVP_CIPHER_CTX_cleanup(ectx);
		    EVP_CIPHER_CTX_cleanup(dctx);
		    EVP_cleanup();
		    free(ectx); free(dctx);
		    ectx = openssl_encrypt_init(opt);
		    dctx = openssl_encrypt_init(opt);
#else
		    gcry_cipher_close(ectx);
		    gcry_cipher_close(dctx);
		    ectx = gcrypt_crypto_init(opt);
		    dctx = gcrypt_crypto_init(opt);
#endif
		}
		FD_ZERO(&rfd);
		FD_SET(net_srv->sockfd, &rfd);
		FD_SET(net_cli->sockfd, &rfd);
		ret = select(net_srv->sockfd + net_cli->sockfd + 1, &rfd,
			     NULL, NULL, NULL);
		if (FD_ISSET(net_srv->sockfd, &rfd)) {
		    if (opt.proto == 0)
			ret = read(net_srv->sockfd, data, PACKET_SIZE * 2);
		    else
			ret = recvfrom(net_srv->sockfd, data,
				       PACKET_SIZE * 2, 0,
				       (struct sockaddr *) &net_srv->addr,
				       &net_srv->len);
		    if (ret <= 0) 
			break;
		    if (opt.server == 0) {
			if (opt.comp == 1) {
			    pkt_in = do_encrypt(ectx, (unsigned char *) data,
					        ret, opt.key.key, opt.key.iv);
			    pkt_out = z_compress(comp.zip, pkt_in.data,
				                 pkt_in.len);
			} else {
			    pkt_out = do_encrypt(ectx, (unsigned char *) data,
					         ret, opt.key.key, opt.key.iv);
			}
		    } else {
			if (opt.comp == 1) {
			    pkt_in = z_uncompress(comp.unzip, data, ret);
			    pkt_out = do_decrypt(dctx,
				                 (unsigned char *) pkt_in.data,
					         pkt_in.len, opt.key.key,
					         opt.key.iv);
			} else {
			    pkt_out = do_decrypt(dctx, (unsigned char *) data,
					         ret, opt.key.key, opt.key.iv);
			}
		    }
		    if (opt.proto == 0)
			ret = write(net_cli->sockfd, pkt_out.data, pkt_out.len);
		    else
			ret = sendto(net_cli->sockfd, pkt_out.data,
				     pkt_out.len, 0,
				     (struct sockaddr *) &net_cli->addr,
				     (socklen_t) sizeof(struct sockaddr));
		    if (opt.comp == 1) {
			free(pkt_in.data);
			free(pkt_out.data);
		    } else {
			free(pkt_out.data);
		    }
		    memset(data, 0x00, sizeof(char *) * PACKET_SIZE);
		}
		if (FD_ISSET(net_cli->sockfd, &rfd)) {
		    if (opt.proto == 0)
			ret = read(net_cli->sockfd, data, PACKET_SIZE * 2);
		    else
			ret = recvfrom(net_cli->sockfd, data,
				       PACKET_SIZE * 2, 0,
				       (struct sockaddr *) &net_cli->addr,
				       &net_cli->len);
		    if (ret <= 0) 
			break;
		    if (opt.server == 0) {
			if (opt.comp == 1) {
			    pkt_in = z_uncompress(comp.unzip, data, ret);
			    pkt_out = do_decrypt(dctx, pkt_in.data, pkt_in.len,
					         opt.key.key, opt.key.iv);
			} else {
			    pkt_out = do_decrypt(dctx, data, ret,
						 opt.key.key, opt.key.iv);
			}
		    } else {
			if (opt.comp == 1) {
			    pkt_in = do_encrypt(ectx, data, ret,
						opt.key.key, opt.key.iv);
			    pkt_out =
				z_compress(comp.zip, pkt_in.data, pkt_in.len);
			} else {
			    pkt_out = do_encrypt(ectx, data, ret,
						 opt.key.key, opt.key.iv);
			}
		    }
		    if (opt.proto == 0)
			ret = write(net_srv->sockfd, pkt_out.data, pkt_out.len);
		    else
			ret = sendto(net_srv->sockfd, pkt_out.data,
				     pkt_out.len, 0,
				     (struct sockaddr *) &net_srv->addr,
				     (socklen_t) sizeof(struct sockaddr));
		    if (opt.comp == 1) {
			free(pkt_in.data);
			free(pkt_out.data);
		    } else {
			free(pkt_out.data);
		    }
		    memset(data, 0x00, sizeof(char *) * PACKET_SIZE);
		}
		if (opt.proto == 1 && opt.comp == 1) {
		    deflateEnd(&comp.zip);
		    inflateEnd(&comp.unzip);
		    comp = z_compress_init(opt);
		}
	    }
	    /* Connection closed */
	    FD_CLR(net_cli->sockfd, &rfd);
	    FD_CLR(net_srv->sockfd, &rfd);
	    if (opt.proto == 0) {
		close(net_cli->sockfd);
		close(net_srv->sockfd);
		free(data);
#ifdef HAVE_OPENSSL
		EVP_CIPHER_CTX_cleanup(ectx);
		EVP_CIPHER_CTX_cleanup(dctx);
		EVP_cleanup();
#else
		gcry_cipher_close(ectx);
		gcry_cipher_close(dctx);
#endif
		if (opt.comp == 1) {
		    deflateEnd(&comp.zip);
		    inflateEnd(&comp.unzip);
		}
		    
		free(net_cli);
		free(net_srv);
		_exit(0);
	    }
	} else {
	    close(net_srv->sockfd);
	}
    }
    return 0;
}
